#!/usr/bin/python3

import contextlib
import os
import subprocess

import pytest

from osbuild.testutil import has_executable

STAGE_NAME = "org.osbuild.btrfs.subvol"


def make_btrfs_disk(tmp_path):
    fake_disk_path = tmp_path / "fake.img"
    with fake_disk_path.open("w") as fp:
        fp.truncate(110 * 1024 * 1024)
    subprocess.run(["mkfs.btrfs", fake_disk_path], check=True)
    return fake_disk_path


@pytest.mark.skipif(os.getuid() != 0, reason="needs root")
@pytest.mark.skipif(not has_executable("mkfs.btrfs"), reason="need mkfs.btrfs")
def test_btrfs_subvol_integration(tmp_path, stage_module):
    fake_disk_path = make_btrfs_disk(tmp_path)
    fake_disk_mnt = tmp_path / "mnt"
    fake_disk_mnt.mkdir()
    with contextlib.ExitStack() as cm:
        subprocess.run(["mount", fake_disk_path, fake_disk_mnt], check=True)
        cm.callback(subprocess.run, ["umount", fake_disk_mnt], check=True)

        paths = {
            "mounts": fake_disk_mnt,
        }
        options = {
            "subvolumes": [
                {"name": "/asubvol"},
                {"name": "/subvols/root"},
            ],
        }
        stage_module.main(paths, options)

        output = subprocess.check_output([
            "btrfs", "subvolume", "list", fake_disk_mnt], encoding="utf-8")
        assert "path asubvol\n" in output
        assert "path subvols/root\n" in output



#!/usr/bin/python3

from unittest.mock import patch

import pytest

STAGE_NAME = "org.osbuild.gzip"


@pytest.mark.parametrize("test_options,expected_level", [
    # our default
    ({}, "-1"),
    # custom settings
    ({"level": "1"}, "-1"),
    ({"level": "6"}, "-6"),
    ({"level": "9"}, "-9"),
])
@patch("subprocess.run")
def test_gzip_compression_default_to_level1(mocked_run, tmp_path, stage_module, test_options, expected_level):
    inp = {
        "file": {
            "path": "/input/file/path",
            "data": {
                "files": {
                    "hash:value": "some-file",
                }
            },
        },
    }
    output = tmp_path
    options = {
        "filename": "out.tar.gz",
    }
    options.update(test_options)

    stage_module.main(inp, output, options)
    assert len(mocked_run.call_args_list) == 1
    args, kwargs = mocked_run.call_args_list[0]
    assert args == (["gzip", "--no-name", "--stdout", expected_level, "/input/file/path/hash:value"],)
    assert kwargs["check"]
    assert kwargs["stdout"].name.endswith("out.tar.gz")



"""EFI boot image utilities

Functions for building EFI boot directory layouts and
FAT filesystem images suitable for ISO / removable-media boot.
"""
import glob
import os
import shutil
import subprocess
import tarfile
import tempfile

from osbuild.util.path import ensure_glob


def find_efi_source_paths(tree):
    """
    This function finds the source EFI paths to copy from /usr/ to the EFI/
    directory. Historically this was just the usr/lib/bootupd/updates/EFI/
    directory (populated by bootupd) or boot/efi/EFI (package mode), but
    as part of BootLoaderUpdatesPhase1 [1] the shim and grub packages started
    delivering files in /usr/lib/efi/(grub|shim)/<version>/EFI/ directly.
    Lets prefer the new paths and fallback to the legacy paths.

    [1] https://fedoraproject.org/wiki/Changes/BootLoaderUpdatesPhase1
    """
    if glob.glob(os.path.join(tree, 'usr/lib/efi/*/*/EFI')):
        # BootLoaderUpdatesPhase1 has been implemented and we should
        # be able to pick up files from usr/lib/efi/(grub|shim)/<version>/EFI/
        # Let's ensure there's only 2 dirs that match (i.e. one for grub
        # one for shim) and return that list.
        return ensure_glob(os.path.join(tree, 'usr/lib/efi/*/*/EFI'), n=2)
    # Legacy paths:
    #   - 'usr/lib/bootupd/updates/EFI' - on Image Mode systems
    #   - 'boot/efi/EFI'                - on Package Mode systems
    for path in ['usr/lib/bootupd/updates/EFI', 'boot/efi/EFI']:
        fullpath = os.path.join(tree, path)
        if os.path.exists(fullpath):
            return [fullpath]
    raise ValueError(f'No EFI source paths found in {tree}')


def find_efi_vendor_dir_name(efidir=None, source_tree=None):
    """
    Searches either an EFI directory structure (efidir) or a source tree
    (source_tree) for an EFI vendor directory name (usually fedora,
    redhat, or centos). If source_tree is given the efidir will be found
    first using find_efi_source_paths().
    """
    if source_tree and efidir:
        raise ValueError("find_efi_vendor_dir_name: cannot pass both efidir and source_tree")
    if not source_tree and not efidir:
        raise ValueError("find_efi_vendor_dir_name: must pass either efidir or source_tree")
    # If the user provided a source_tree and not an efidir then
    # we must first find the efidir within the source_tree.
    if source_tree:
        # To find the string for the vendor ID we only need to inspect one
        # of the source EFI directories. Grab one here.
        efidir = find_efi_source_paths(source_tree)[0]
    # Find name of vendor directory for this distro. i.e. "fedora" or "redhat".
    dirs = [n for n in os.listdir(efidir) if n != "BOOT"]
    if len(dirs) != 1:
        raise ValueError(f"did not find exactly one EFI vendor ID: {dirs}")
    vendor_id = dirs[0]
    return vendor_id


def mkefidir(efidir, source_tree):
    """
    Searches the source_tree for files used for EFI booting and copies
    these files into the specified efidir.
    """
    for path in find_efi_source_paths(source_tree):
        shutil.copytree(path, efidir, dirs_exist_ok=True)

    vendor_id = find_efi_vendor_dir_name(efidir=efidir)

    # Delete fallback and its CSV file.  Its purpose is to create
    # EFI boot variables, which we don't want when booting from
    # removable media.
    #
    # A future shim release will merge fallback.efi into the main
    # shim binary and enable the fallback behavior when the CSV
    # exists.  But for now, fail if fallback.efi is missing.
    for path in ensure_glob(os.path.join(efidir, "BOOT", "fb*.efi")):
        os.unlink(path)
    for path in ensure_glob(os.path.join(efidir, vendor_id, "BOOT*.CSV")):
        os.unlink(path)

    # Drop vendor copies of shim; we already have it in BOOT*.EFI in
    # BOOT
    for path in ensure_glob(os.path.join(efidir, vendor_id, "shim*.efi")):
        os.unlink(path)

    # Consolidate remaining files into BOOT.  shim needs GRUB to be
    # there, and the rest doesn't hurt.
    for path in ensure_glob(os.path.join(efidir, vendor_id, "*")):
        shutil.move(path, os.path.join(efidir, "BOOT"))
    os.rmdir(os.path.join(efidir, vendor_id))


def mkefiboot(efidir, output_efiboot_img, loop_client):
    """
    Creates an efi boot image file at output_efiboot_img (usually named efiboot.img),
    which is required for EFI booting. This is a fat32 formatted filesystem that
    contains all the files needed for EFI boot.
    """

    # In restrictive environments, setgid, setuid and ownership changes
    # may be restricted. This sets the file ownership to root and
    # removes the setgid and setuid bits in the tarball.
    def strip_tar(tarinfo):
        tarinfo.uid = 0
        tarinfo.gid = 0
        if tarinfo.isdir():
            tarinfo.mode = 0o755
        elif tarinfo.isfile():
            tarinfo.mode = 0o0644
        return tarinfo

    # Install binaries from the efidir
    # Manually construct the tarball to ensure proper permissions and ownership
    efitarfile = tempfile.NamedTemporaryFile(suffix=".tar")
    with tarfile.open(efitarfile.name, "w:", dereference=True) as tar:
        tar.add(efidir, arcname="/EFI", filter=strip_tar)

    # Create the efiboot image file. Determine the size we should make
    # it by taking the tarball size and adding 2MiB for fs overhead.
    size = os.path.getsize(efitarfile.name) + 2 * 1024 * 1024
    # Align to 512 byte boundary. Some firmeware (like nvidiabluefield)
    # derives block device sector size from image size; unaligned sizes
    # produce sector size 1 which GRUB cannot handle
    size = (size + 511) & ~511
    with open(output_efiboot_img, "wb") as out:
        out.truncate(size)

    # Make loopback device; mkfs; populate with files
    with loop_client.device(output_efiboot_img) as loopdev:
        # On RHEL 8, when booting from a disk device (rather than a CD),
        # https://github.com/systemd/systemd/issues/14408 causes the
        # hybrid ESP to race with the ISO9660 filesystem for the
        # /dev/disk/by-label symlink unless the ESP has its own label,
        # so set EFI-SYSTEM for consistency with the metal image.
        # This should not be needed on Fedora or RHEL 9, but seems like
        # a good thing to do anyway.
        label = 'EFI-SYSTEM'
        # NOTE: the arguments to mkfs here match how virt-make-fs calls mkfs
        subprocess.check_call(['mkfs', '-t', 'vfat', '-I', '--mbr=n', '-n', label, loopdev])
        with tempfile.TemporaryDirectory() as d:
            try:
                subprocess.check_call(['mount', '-o', 'utf8', loopdev, d])
                subprocess.check_call(['tar', '-C', d, '-xf', efitarfile.name])
            finally:
                subprocess.check_call(['umount', d])



#!/usr/bin/python3

from unittest.mock import patch

import pytest

from osbuild.testutil import (
    assert_jsonschema_error_contains,
    make_fake_tree,
)

STAGE_NAME = "org.osbuild.users"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({"users": {"!invalid-name": {}}}, "'!invalid-name' does not match any of the regex"),
    ({"users": {"foo": {"home": 0}}}, "0 is not of type 'string'"),
    # good
    ({}, ""),
    ({"users": {"foo": {}}}, ""),
    ({"users": {"foo.bar.baz": {}}}, ""),
])
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {},
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


TEST_CASES = [
    # user_opts,expected commandline args
    ({}, []),
    ({"expiredate": "12345"}, ["--expiredate", "12345"]),
]


@pytest.mark.parametrize("user_opts,expected_args", TEST_CASES)
@patch("subprocess.run")
def test_users_happy(mocked_run, tmp_path, stage_module, user_opts, expected_args):
    make_fake_tree(tmp_path, {
        "/etc/passwd": "",
    })

    options = {
        "users": {
            "foo": {},
        }
    }
    options["users"]["foo"].update(user_opts)

    stage_module.main(tmp_path.as_posix(), options)

    assert len(mocked_run.call_args_list) == 1
    args, kwargs = mocked_run.call_args_list[0]
    assert args[0] == ["useradd", "--root", tmp_path.as_posix()] + expected_args + ["foo"]
    assert kwargs.get("check")


@patch("subprocess.run")
@pytest.mark.parametrize("user_opts,expected_args", TEST_CASES)
def test_users_mock_bin(mocked_run, tmp_path, stage_module, user_opts, expected_args):
    make_fake_tree(tmp_path, {
        "/etc/passwd": "",
    })

    options = {
        "users": {
            "foo": {},
        }
    }
    options["users"]["foo"].update(user_opts)

    stage_module.main(tmp_path.as_posix(), options)

    assert len(mocked_run.call_args_list) == 1
    args, kwargs = mocked_run.call_args_list[0]
    assert args[0] == ["useradd", "--root", tmp_path.as_posix()] + expected_args + ["foo"]
    assert kwargs.get("check")


# separate test right now as it results in two binaries being called
# (adduser,passwd) which our parameter tests cannot do yet

@patch("subprocess.run")
def test_users_with_password_reset_none(mocked_run, tmp_path, stage_module):
    make_fake_tree(tmp_path, {
        "/etc/passwd": "",
    })

    options = {
        "users": {
            "foo": {
            },
        }
    }

    stage_module.main(tmp_path.as_posix(), options)

    assert len(mocked_run.call_args_list) == 1
    args, kwargs = mocked_run.call_args_list[0]
    assert args[0] == ["useradd", "--root", tmp_path.as_posix(), "foo"]
    assert kwargs.get("check")

# separate test right now as it results in two binaries being called
# (adduser,passwd) which our parameter tests cannot do yet


@patch("subprocess.run")
def test_users_with_password_reset_false(mocked_run, tmp_path, stage_module):
    make_fake_tree(tmp_path, {
        "/etc/passwd": "",
    })

    options = {
        "users": {
            "foo": {
                "force_password_reset": False,
            },
        }
    }

    stage_module.main(tmp_path.as_posix(), options)

    assert len(mocked_run.call_args_list) == 1
    args, kwargs = mocked_run.call_args_list[0]  # chroot is the 4th call
    assert args[0] == ["useradd", "--root", tmp_path.as_posix(), "foo"]
    assert kwargs.get("check")

# separate test right now as it results in two binaries being called
# (adduser,passwd) which our parameter tests cannot do yet


@patch("subprocess.run")
def test_users_with_password_reset_true(mocked_run, tmp_path, stage_module):
    make_fake_tree(tmp_path, {
        "/etc/passwd": "",
    })

    options = {
        "users": {
            "foo": {
                "force_password_reset": True,
            },
        }
    }

    stage_module.main(tmp_path.as_posix(), options)

    assert len(mocked_run.call_args_list) == 2

    args, kwargs = mocked_run.call_args_list[0]
    assert args[0] == ["useradd", "--root", tmp_path.as_posix(), "foo"]
    assert kwargs.get("check")

    args, kwargs = mocked_run.call_args_list[1]
    assert args[0] == ["passwd", "--root", tmp_path.as_posix(), "--expire", "foo"]
    assert kwargs.get("check")



"""OS-Release Information

This module implements handlers for the `/etc/os-release` type of files. The
related documentation can be found in `os-release(5)`.
"""

import os
import shlex

# The default paths where os-release is located, as per os-release(5)
DEFAULT_PATHS = [
    "/etc/os-release",
    "/usr/lib/os-release"
]


def parse_files(*paths):
    """Read Operating System Information from `os-release`

    This creates a dictionary with information describing the running operating
    system. It reads the information from the path array provided as `paths`.
    The first available file takes precedence. It must be formatted according
    to the rules in `os-release(5)`.
    """
    osrelease = {}

    path = next((p for p in paths if os.path.exists(p)), None)
    if path:
        with open(path, encoding="utf8") as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                if line[0] == "#":
                    continue
                key, value = line.split("=", 1)
                split_value = shlex.split(value)
                if not split_value:
                    raise ValueError(f"Key '{key}' has an empty value")
                if len(split_value) > 1:
                    raise ValueError(f"Key '{key}' has more than one token: {value}")
                osrelease[key] = split_value[0]

    return osrelease


def describe_os(*paths):
    """Read the Operating System Description from `os-release`

    This creates a string describing the running operating-system name and
    version. It uses `parse_files()` underneath to acquire the requested
    information.

    The returned string uses the format `${ID}${VERSION_ID}` with all dots
    stripped.
    """
    osrelease = parse_files(*paths)

    # Fetch `ID` and `VERSION_ID`. Defaults are defined in `os-release(5)`.
    osrelease_id = osrelease.get("ID", "linux")
    osrelease_version_id = osrelease.get("VERSION_ID", "")

    return osrelease_id + osrelease_version_id.replace(".", "")



#!/usr/bin/python3
import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.fix-bls"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({"unknown": "property"}, "Additional properties are not allowed ('unknown' was unexpected)"),
    # good
    ({}, ""),
    ({"prefix": "/foo/bar"}, ""),
])
def test_fix_bls_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


def test_fix_bls_files_no_prefix(tmp_path, stage_module):
    treepath = tmp_path / "tree"

    blspath = treepath / "boot/loader/entries"
    blspath.mkdir(parents=True)

    conf0path = blspath / "dummy0.conf"
    conf0path.write_text("linux  /run/osbuild/tree/boot/vmlinuz0\ninitrd  /run/osbuild/tree/boot/initramfs0\n")

    conf1path = blspath / "dummy1.conf"
    conf1path.write_text("linux  /run/osbuild/tree/boot/vmlinuz1\ninitrd  /run/osbuild/tree/boot/initramfs1\n")

    conf2path = blspath / "dummy2.conf"
    conf2path.write_text("linux  /boot/vmlinuz2\ninitrd  /boot/initramfs2\n")

    options = {}

    stage_module.main(treepath, options)

    assert conf0path.read_text() == "linux  /boot/vmlinuz0\ninitrd  /boot/initramfs0\n"
    assert conf1path.read_text() == "linux  /boot/vmlinuz1\ninitrd  /boot/initramfs1\n"
    assert conf2path.read_text() == "linux  /boot/vmlinuz2\ninitrd  /boot/initramfs2\n"


def test_fix_bls_files_with_prefix(tmp_path, stage_module):
    treepath = tmp_path / "tree"

    blspath = treepath / "boot/loader/entries"
    blspath.mkdir(parents=True)

    conf0path = blspath / "dummy0.conf"
    conf0path.write_text("linux  /run/osbuild/tree/boot/vmlinuz0\ninitrd  /run/osbuild/tree/boot/initramfs0\n")

    conf1path = blspath / "dummy1.conf"
    conf1path.write_text("linux  /run/osbuild/tree/boot/vmlinuz1\ninitrd  /run/osbuild/tree/boot/initramfs1\n")

    conf2path = blspath / "dummy2.conf"
    conf2path.write_text("linux  /boot/vmlinuz2\ninitrd  /boot/initramfs2\n")

    options = {"prefix": ""}

    stage_module.main(treepath, options)

    assert conf0path.read_text() == "linux  /vmlinuz0\ninitrd  /initramfs0\n"
    assert conf1path.read_text() == "linux  /vmlinuz1\ninitrd  /initramfs1\n"
    assert conf2path.read_text() == "linux  /boot/vmlinuz2\ninitrd  /boot/initramfs2\n"


def test_fix_bls_files_with_prefix_including_boot(tmp_path, stage_module):
    treepath = tmp_path / "tree"

    blspath = treepath / "boot/loader/entries"
    blspath.mkdir(parents=True)

    conf0path = blspath / "dummy0.conf"
    conf0path.write_text("linux  /run/osbuild/tree/boot/vmlinuz0\ninitrd  /run/osbuild/tree/boot/initramfs0\n")

    conf1path = blspath / "dummy1.conf"
    conf1path.write_text("linux  /run/osbuild/tree/boot/vmlinuz1\ninitrd  /run/osbuild/tree/boot/initramfs1\n")

    conf2path = blspath / "dummy2.conf"
    conf2path.write_text("linux  /boot/vmlinuz2\ninitrd  /boot/initramfs2\n")

    options = {"prefix": "", "require_boot_prefix": False}

    stage_module.main(treepath, options)

    assert conf0path.read_text() == "linux  /vmlinuz0\ninitrd  /initramfs0\n"
    assert conf1path.read_text() == "linux  /vmlinuz1\ninitrd  /initramfs1\n"
    assert conf2path.read_text() == "linux  /vmlinuz2\ninitrd  /initramfs2\n"



#!/usr/bin/python3

import contextlib
import os
import subprocess

import pytest  # type: ignore

from osbuild.testutil import has_executable

STAGE_NAME = "org.osbuild.mkdir"


def test_mkdir(tmp_path, stage_module):
    tree = tmp_path / "tree"
    tree.mkdir()

    options = {
        "paths": [
            {"path": "/fake_dir"},
            {"path": "fake_relative_dir"}
        ]
    }
    args = {
        "tree": f"{tree}",
        "options": options
    }
    stage_module.main(args)
    assert (tree / "fake_dir").exists()
    assert (tree / "fake_relative_dir").exists()


def test_mkdir_on_a_tree(tmp_path, stage_module):
    tree = tmp_path / "tree"
    tree.mkdir()

    options = {
        "paths": [
            {
                "path": "tree:///fake_parent/fake_dir",
                "parents": 1
            }
        ]
    }
    args = {
        "tree": f"{tree}",
        "options": options
    }
    stage_module.main(args)
    assert (tree / "fake_parent/fake_dir").exists()


@pytest.mark.skipif(os.getuid() != 0, reason="needs root")
@pytest.mark.skipif(not has_executable("mkfs.ext4"), reason="need mkfs.ext4")
def test_mkdir_on_a_mount(tmp_path, stage_module):
    tree = tmp_path / "tree"
    tree.mkdir()

    # Create fake EXT4 disk image
    fake_disk_path = tmp_path / "fake.img"
    with fake_disk_path.open("w") as fp:
        fp.truncate(10 * 1024 * 1024)
    subprocess.run(
        ["mkfs.ext4", os.fspath(fake_disk_path)], check=True)

    fake_disk_mnt = tmp_path / "mounts"
    fake_disk_mnt.mkdir()

    with contextlib.ExitStack() as cm:
        subprocess.run(["mount", fake_disk_path, fake_disk_mnt], check=True)
        cm.callback(subprocess.run, ["umount", fake_disk_mnt], check=True)

        options = {
            "paths": [
                {
                    "path": "mount:///fake_parent/fake_dir",
                    "parents": 1
                }
            ]
        }
        args = {
            "tree": f"{tree}",
            "options": options,
            "paths": {
                "mounts": fake_disk_mnt,
            }
        }
        stage_module.main(args)
        assert (fake_disk_mnt / "fake_parent/fake_dir").exists()



#!/usr/bin/python3
import os
import subprocess
from unittest.mock import call, patch

import pytest

STAGE_NAME = "org.osbuild.grub2.inst"


# pylint: disable=inconsistent-return-statements
def find_grub_mkimage():
    for exe in ["grub2-mkimage", "grub-mkimage"]:
        try:
            subprocess.check_call([exe, "-V"])
            return exe
        except FileNotFoundError:
            pass
    pytest.skip("cannot find grub{,2}-mkimage")


def test_grub2_partition(tmp_path, stage_module):
    treedir = tmp_path / "tree"
    os.makedirs(treedir, exist_ok=True)
    with open(treedir / "disk.img", "wb") as f:
        f.write(b"Just testing")

    options = {
        "filename": "disk.img",
        "platform": "i386-pc",
        "location": 0,
        "core": {
            "binary": find_grub_mkimage(),
            "type": "mkimage",
            "partlabel": "gpt",
            "filesystem": "ext4",
        },
        "prefix": {
            "type": "partition",
            "partlabel": "gpt",
            "number": 0,
            "path": "/boot/",
        },
    }
    stage_module.main(treedir, options)

    # Make sure the test string is overwritten
    with open(treedir / "disk.img", "rb") as f:
        msg = f.read(12)
        assert msg != b"Just testing"


def test_grub2_iso9660(tmp_path, stage_module):
    treedir = tmp_path / "tree"
    os.makedirs(treedir, exist_ok=True)

    options = {
        "filename": "eltorito.img",
        "platform": "i386-pc",
        "core": {
            "binary": find_grub_mkimage(),
            "type": "mkimage",
            "partlabel": "gpt",
            "filesystem": "iso9660",
        },
        "prefix": {
            "path": "/boot/grub2/",
        },
    }
    stage_module.main(treedir, options)
    assert os.path.exists(treedir / "eltorito.img")


@patch("subprocess.run")
def test_grub2_partition_mocked(mocked_run, tmp_path, stage_module):
    options = {
        "filename": "disk.img",
        "platform": "some-platform",
        "location": 0,
        "core": {
            "type": "mkimage",
            "partlabel": "gpt",
            "filesystem": "ext4",
        },
        "prefix": {
            "type": "partition",
            "partlabel": "gpt",
            "number": 0,
            "path": "/boot/",
        },
    }

    output_path = ""

    def mock_side_effect(args, **_):
        """
        Intercept the args to subprocess.run() to leak the random output path and create the file that's expected by the
        shutil.copy() in the stage's main().
        """
        nonlocal output_path
        output_path = args[args.index("--output") + 1]
        with open(output_path, "wb") as fake_image_file:
            fake_image_file.write(b"I am a grub core image")

    mocked_run.side_effect = mock_side_effect

    stage_module.main(tmp_path, options)

    assert os.path.basename(output_path) == "grub2-core.img"
    assert mocked_run.call_args_list == [
        call(["grub2-mkimage", "--verbose",
              "--directory", "/usr/lib/grub/some-platform",
              "--prefix", "(,gpt1)/boot/",
              "--format", "some-platform",
              "--compression", "auto",
              "--output", output_path,
              "part_gpt", "ext2"], check=True),
    ]



import collections
import contextlib
import glob
import json
import os
import re
import subprocess
import sys
import tempfile
import typing

# pylint doesn't understand the string-annotation below
from typing import Any, Dict, List, Tuple  # pylint: disable=unused-import

from osbuild.util.rhsm import Subscriptions

from .types import PathLike


class Param:
    """rpm-ostree Treefile parameter"""

    def __init__(self, value_type, mandatory=False):
        self.type = value_type
        self.mandatory = mandatory

    def check(self, value):
        origin = getattr(self.type, "__origin__", None)
        if origin:
            self.typecheck(value, origin)
            if origin is list or origin is typing.List:
                self.check_list(value, self.type)
            else:
                raise NotImplementedError(origin)
        else:
            self.typecheck(value, self.type)

    @staticmethod
    def check_list(value, tp):
        inner = tp.__args__
        for x in value:
            Param.typecheck(x, inner)

    @staticmethod
    def typecheck(value, tp):
        if isinstance(value, tp):
            return
        raise ValueError(f"{value} is not of {tp}")


class Treefile:
    """Representation of an rpm-ostree Treefile

    The following parameters are currently supported,
    presented together with the rpm-ostree compose
    phase that they are used in.
      - ref: commit
      - repos: install
      - selinux: install, postprocess, commit
      - boot-location: postprocess
      - etc-group-members: postprocess
      - machineid-compat
      - selinux-label-version: commit

    NB: 'ref' and 'repos' are mandatory and must be
    present, even if they are not used in the given
    phase; they therefore have defaults preset.
    """

    parameters = {
        "ref": Param(str, True),
        "repos": Param(List[str], True),
        "selinux": Param(bool),
        "boot-location": Param(str),
        "etc-group-members": Param(List[str]),
        "machineid-compat": Param(bool),
        "initramfs-args": Param(List[str]),
        "selinux-label-version": Param(int),
    }

    def __init__(self):
        self._data = {}
        self["ref"] = "osbuild/devel"
        self["repos"] = ["osbuild"]

    def __getitem__(self, key):
        param = self.parameters.get(key)
        if not param:
            raise ValueError(f"Unknown param: {key}")
        return self._data[key]

    def __setitem__(self, key, value):
        param = self.parameters.get(key)
        if not param:
            raise ValueError(f"Unknown param: {key}")
        param.check(value)
        self._data[key] = value

    def dumps(self):
        return json.dumps(self._data)

    def dump(self, fp):
        return json.dump(self._data, fp)

    @contextlib.contextmanager
    def as_tmp_file(self):
        name = None
        try:
            fd, name = tempfile.mkstemp(suffix=".json",
                                        text=True)

            with os.fdopen(fd, "w+", encoding="utf8") as f:
                self.dump(f)

            yield name
        finally:
            if name:
                os.unlink(name)


def setup_remote(repo, name, remote):
    """Configure an OSTree remote in a given repo"""

    url = remote["url"]
    gpg = remote.get("gpgkeys", [])

    remote_add_args = []
    if not gpg:
        remote_add_args = ["--no-gpg-verify"]

    if "contenturl" in remote:
        remote_add_args.append(f"--contenturl={remote['contenturl']}")

    if remote.get("secrets", {}).get("name") == "org.osbuild.rhsm.consumer":
        secrets = Subscriptions.get_consumer_secrets()
        remote_add_args.append(f"--set=tls-client-key-path={secrets['consumer_key']}")
        remote_add_args.append(f"--set=tls-client-cert-path={secrets['consumer_cert']}")
    elif remote.get("secrets", {}).get("name") == "org.osbuild.mtls":
        tlsca = os.getenv("OSBUILD_SOURCES_OSTREE_SSL_CA_CERT")
        if tlsca:
            remote_add_args.append(f"--set=tls-ca-path={tlsca}")

        tlscert = os.getenv("OSBUILD_SOURCES_OSTREE_SSL_CLIENT_CERT")
        if tlscert:
            remote_add_args.append(f"--set=tls-client-cert-path={tlscert}")

        tlskey = os.getenv("OSBUILD_SOURCES_OSTREE_SSL_CLIENT_KEY")
        if tlskey:
            remote_add_args.append(f"--set=tls-client-key-path={tlskey}")

        proxy = os.getenv("OSBUILD_SOURCES_OSTREE_PROXY")
        if proxy:
            remote_add_args.append(f"--set=proxy={proxy}")

        # Insecure mode is meant for development only
        insecure = os.getenv("OSBUILD_SOURCES_OSTREE_INSECURE")
        if insecure and insecure.lower() in ["true", "yes", "1"]:
            remote_add_args.append("--set=tls-permissive=true")

    cli("remote", "add", name, url,
        *remote_add_args, repo=repo)

    for key in gpg:
        cli("remote", "gpg-import", "--stdin",
            name, repo=repo, _input=key)


def rev_parse(repo: PathLike, ref: str) -> str:
    """Resolve an OSTree reference `ref` in the repository at `repo`"""

    repo = os.fspath(repo)

    if isinstance(repo, bytes):
        repo = repo.decode("utf8")

    r = subprocess.run(["ostree", "rev-parse", ref, f"--repo={repo}"],
                       encoding="utf8",
                       stdout=subprocess.PIPE,
                       stderr=subprocess.STDOUT,
                       check=False)

    msg = r.stdout.strip()
    if r.returncode != 0:
        raise RuntimeError(msg)

    return msg


def show(repo: PathLike, checksum: str) -> str:
    """Show the metada of an OSTree object pointed by `checksum` in the repository at `repo`"""

    repo = os.fspath(repo)

    if isinstance(repo, bytes):
        repo = repo.decode("utf8")

    r = subprocess.run(["ostree", "show", f"--repo={repo}", checksum],
                       encoding="utf8",
                       stdout=subprocess.PIPE,
                       stderr=subprocess.STDOUT,
                       check=False)

    msg = r.stdout.strip()
    if r.returncode != 0:
        raise RuntimeError(msg)

    return msg


def pull_local(source_repo: PathLike, target_repo: PathLike, remote: str, ref: str):
    """Run ostree-pull local to copy commits around"""

    extra_args = []
    if remote:
        extra_args.append(f'--remote={remote}')

    cli("pull-local", source_repo, ref,
        *extra_args,
        repo=target_repo)


def cli(*args, _input=None, **kwargs):
    """Thin wrapper for running the ostree CLI"""
    args = list(args) + [f'--{k}={v}' for k, v in kwargs.items()]
    print("ostree " + " ".join(args), file=sys.stderr)
    return subprocess.run(["ostree"] + args,
                          encoding="utf8",
                          stdout=subprocess.PIPE,
                          input=_input,
                          check=True)


def parse_input_commits(commits):
    """Parse ostree input commits and return the repo path and refs specified"""
    data = commits["data"]
    refs = data["refs"]
    assert refs, "Need at least one commit"
    return commits["path"], data["refs"]


def default_boot_path(root: PathLike) -> str:
    """Find the default ostree boot path and return it

    eg. /ostree/boot.1/default/b6cdf47cafd171e003c001802f1829ad39e20f7177d6e27401aba73efb71be22/0
    """
    filenames = glob.glob(os.path.join(root, 'ostree/boot.?/*/*/*'))
    if len(filenames) < 1:
        raise ValueError("Could not find boot path")
    if len(filenames) > 1:
        raise ValueError(f"More than one boot path found: {filenames}")
    return "/" + os.path.relpath(filenames[0], root)


def parse_deployment_option(root: PathLike, deployment: Dict) -> Tuple[str, str, str]:
    """Parse the deployment option and return the osname, ref, and serial

    The `deployment` arg contains the following sub fields:
    - osname: Name of the stateroot used in the deployment (ie. fedora-coreos)
    - ref: OStree ref to used for the deployment (ie. fedora/aarch64/coreos/next)
    - serial: The deployment serial (ie. 0)
    - default: Boolean to determine whether the default ostree deployment should be used
    """

    default_deployment = deployment.get("default")
    if default_deployment:
        filenames = glob.glob(os.path.join(root, 'ostree/deploy/*/deploy/*.0'))
        if len(filenames) < 1:
            raise ValueError("Could not find deployment")
        if len(filenames) > 1:
            raise ValueError(f"More than one deployment found: {filenames}")

        # We pick up the osname, commit, and serial from the filesystem
        # here. We'll return the detected commit as the ref in this
        # since it's a valid substitute for all subsequent uses in
        # the code base.
        f = re.search("/ostree/deploy/(.*)/deploy/(.*)\\.([0-9])", filenames[0])
        if not f:
            raise ValueError("cannot find ostree deployment in {filenames[0]}")
        osname = f.group(1)
        commit = f.group(2)
        serial = f.group(3)
        return osname, commit, serial

    osname = deployment["osname"]
    ref = deployment["ref"]
    serial = deployment.get("serial", 0)
    return osname, ref, serial


def deployment_path(root: PathLike, osname: str = "", ref: str = "", serial: int = 0):
    """Return the path to a deployment given the parameters"""

    base = os.path.join(root, "ostree")

    repo = os.path.join(base, "repo")
    stateroot = os.path.join(base, "deploy", osname)

    commit = rev_parse(repo, ref)
    sysroot = f"{stateroot}/deploy/{commit}.{serial}"

    return sysroot


def parse_origin(origin: PathLike):
    """Parse the origin file and return the deployment type and imgref

    Example container case: container-image-reference=ostree-remote-image:fedora:docker://quay.io/fedora/fedora-coreos:stable
    Example ostree commit case: refspec=fedora:fedora/x86_64/coreos/stable
    """
    deploy_type = ""
    imgref = ""
    with open(origin, "r", encoding="utf8") as f:
        for line in f:
            separated_line = line.split("=")
            if separated_line[0] == "container-image-reference":
                deploy_type = "container"
                imgref = separated_line[1].rstrip()
                break
            if separated_line[0] == "refspec":
                deploy_type = "ostree_commit"
                imgref = separated_line[1].rstrip()
                break

    if deploy_type == "":
        raise ValueError("Could not find 'container-image-reference' or 'refspec' in origin file")
    if imgref == "":
        raise ValueError("Could not find imgref in origin file")

    return deploy_type, imgref


class PasswdLike:
    """Representation of a file with structure like /etc/passwd

    If each line in a file contains a key-value pair separated by the
    first colon on the line, it can be considered "passwd"-like. This
    class can parse the the list, manipulate it, and export it to file
    again.
    """

    def __init__(self):
        """Initialize an empty PasswdLike object"""
        self.db = {}

    @classmethod
    def from_file(cls, path: PathLike, allow_missing_file: bool = False):
        """Initialize a PasswdLike object from an existing file"""
        ret = cls()
        if allow_missing_file:
            if not os.path.isfile(path):
                return ret

        with open(path, "r", encoding="utf8") as p:
            ret.db = cls._passwd_lines_to_dict(p.readlines())
        return ret

    def merge_with_file(self, path: PathLike, allow_missing_file: bool = False):
        """Extend the database with entries from another file"""
        if allow_missing_file:
            if not os.path.isfile(path):
                return

        with open(path, "r", encoding="utf8") as p:
            additional_passwd_dict = self._passwd_lines_to_dict(p.readlines())
            for name, passwd_line in additional_passwd_dict.items():
                if name not in self.db:
                    self.db[name] = passwd_line

    def dump_to_file(self, path: PathLike):
        """Write the current database to a file"""
        with open(path, "w", encoding="utf8") as p:
            p.writelines(list(self.db.values()))

    @staticmethod
    def _passwd_lines_to_dict(lines):
        """Take a list of passwd lines and produce a "name": "line" dictionary"""
        return {line.split(':')[0]: line for line in lines}


class SubIdsDB:
    """Represention of subordinate Ids database

    Class to represent a mapping of a user name to subordinate ids,
    like `/etc/subgid` and `/etc/subuid`.
    """

    def __init__(self) -> None:
        self.db: 'collections.OrderedDict[str, Any]' = collections.OrderedDict()

    def read(self, fp) -> int:
        idx = 0
        for idx, line in enumerate(fp.readlines()):
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            comps = line.split(":")
            if len(comps) != 3:
                print(f"WARNING: invalid line `{line}`", file=sys.stderr)
                continue
            name, uid, count = comps
            self.db[name] = (uid, count)
        return idx

    def dumps(self) -> str:
        """Dump the database to a string"""
        data = "\n".join([
            f"{name}:{uid}:{count}\n"
            for name, (uid, count) in self.db.items()
        ])

        return data

    def read_from(self, path: PathLike) -> int:
        """Read a file and add the entries to the database"""
        with open(path, "r", encoding="utf8") as f:
            return self.read(f)

    def write_to(self, path: PathLike) -> None:
        """Write the database to a file"""
        data = self.dumps()
        with open(path, "w", encoding="utf8") as f:
            f.write(data)

    def __bool__(self) -> bool:
        return bool(self.db)



"""Entrypoints for osbuild

This module contains the application and API entrypoints of `osbuild`, the
command-line-interface to osbuild. The `osbuild_cli()` entrypoint can be safely
used from tests to run the cli.
"""


import argparse
import json
import os
import sys
import typing
from typing import List

import osbuild
import osbuild.meta
import osbuild.monitor
from osbuild.meta import ValidationResult
from osbuild.objectstore import ObjectStore
from osbuild.pipeline import Manifest
from osbuild.util.parsing import parse_size
from osbuild.util.term import fmt as vt


def parse_manifest(path: str) -> dict:
    if path == "-":
        manifest = json.load(sys.stdin)
    else:
        with open(path, encoding="utf8") as f:
            manifest = json.load(f)

    return manifest


def show_validation(result: ValidationResult, name: str) -> None:
    if name == "-":
        name = "<stdin>"

    print(f"{vt.bold}{name}{vt.reset} ", end='')

    if result:
        print(f"is {vt.bold}{vt.green}valid{vt.reset}")
        return

    print(f"has {vt.bold}{vt.red}errors{vt.reset}:")
    print("")

    for error in result:
        print(f"{vt.bold}{error.id}{vt.reset}:")
        print(f"  {error.message}\n")


def export(name_or_id: str, output_directory: str, store: ObjectStore, manifest: Manifest) -> None:
    pipeline = manifest[name_or_id]
    obj = store.get(pipeline.id)
    dest = os.path.join(output_directory, name_or_id)

    skip_preserve_owner = \
        os.getenv("OSBUILD_EXPORT_FORCE_NO_PRESERVE_OWNER") == "1"
    os.makedirs(dest, exist_ok=True)
    obj.export(dest, skip_preserve_owner=skip_preserve_owner)


@typing.no_type_check  # see https://github.com/python/typeshed/issues/3107
def parse_arguments(sys_argv: List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="osbuild",
                                     description="Build operating system images")

    parser.add_argument("manifest_path", metavar="MANIFEST",
                        help="json file containing the manifest that should be built, or a '-' to read from stdin")
    parser.add_argument("--cache", "--store", metavar="DIRECTORY", type=os.path.abspath,
                        default=".osbuild",
                        help="directory where sources and intermediary os trees are stored")
    parser.add_argument("-l", "--libdir", metavar="DIRECTORY", type=os.path.abspath, default="/usr/lib/osbuild",
                        help="directory containing stages, assemblers, and the osbuild library")
    parser.add_argument("--cache-max-size", metavar="SIZE", type=parse_size, default=None,
                        help="maximum size of the cache (bytes) or 'unlimited' for no restriction")
    parser.add_argument(
        "--checkpoint",
        metavar="ID",
        action="append",
        type=str,
        default=None,
        help="stage to commit to the object store during build (can be passed multiple times), accepts globs")
    parser.add_argument("--export", metavar="ID", action="append", type=str, default=[],
                        help="object to export, can be passed multiple times")
    parser.add_argument("--in-vm", metavar="ID", action="append", type=str, default=[],
                        help="Run a pipeline in a VM")
    parser.add_argument("--json", action="store_true",
                        help="output results in JSON format")
    parser.add_argument("--output-directory", metavar="DIRECTORY", type=os.path.abspath,
                        help="directory where result objects are stored")
    parser.add_argument("--inspect", action="store_true",
                        help="return the manifest in JSON format including all the ids")
    parser.add_argument("--monitor", metavar="NAME", default=None,
                        help="name of the monitor to be used")
    parser.add_argument("--monitor-fd", metavar="FD", type=int, default=sys.stdout.fileno(),
                        help="file descriptor to be used for the monitor")
    parser.add_argument("--stage-timeout", type=int, default=None,
                        help="set the maximal time (in seconds) each stage is allowed to run")
    parser.add_argument("--version", action="version",
                        help="return the version of osbuild",
                        version="%(prog)s " + osbuild.__version__)
    # nargs='?' const='*' means `--break` is equivalent to `--break=*`
    parser.add_argument("--break", dest='debug_break', type=str, nargs='?', const='*',
                        help="open debug shell when executing stage. Accepts stage name or id or * (for all)")
    parser.add_argument("--quiet", "-q", action="store_true",
                        help="suppress normal output")

    return parser.parse_args(sys_argv[1:])


# pylint: disable=too-many-branches,too-many-return-statements,too-many-statements
def osbuild_cli() -> int:
    args = parse_arguments(sys.argv)
    desc = parse_manifest(args.manifest_path)

    index = osbuild.meta.Index(args.libdir)

    # detect the format from the manifest description
    info = index.detect_format_info(desc)
    if not info:
        print("Unsupported manifest format")
        return 2
    fmt = info.module

    # first thing is validation of the manifest
    res = fmt.validate(desc, index)
    if not res:
        if args.json or args.inspect:
            json.dump(res.as_dict(), sys.stdout)
            sys.stdout.write("\n")
        else:
            show_validation(res, args.manifest_path)
        return 2

    manifest = fmt.load(desc, index)

    exports = set(args.export)
    unresolved = [e for e in exports if e not in manifest]
    if unresolved:
        available = list(manifest.pipelines.keys())
        for name in unresolved:
            print(f"Export {vt.bold}{name}{vt.reset} not found in {available}")
        print(f"{vt.reset}{vt.bold}{vt.red}Failed{vt.reset}")
        return 1

    if args.checkpoint:
        marked = manifest.mark_checkpoints(args.checkpoint)
        if not marked:
            print("No checkpoints matched provided patterns!")
            print(f"{vt.reset}{vt.bold}{vt.red}Failed{vt.reset}")
            return 1
    else:
        marked = set()

    if args.inspect:
        result = fmt.describe(manifest, with_id=True)
        json.dump(result, sys.stdout)
        sys.stdout.write("\n")
        return 0

    output_directory = args.output_directory

    if exports and not output_directory:
        print("Need --output-directory for --export")
        return 1

    in_vm = set(args.in_vm)

    monitor_name = args.monitor
    if not monitor_name:
        monitor_name = "NullMonitor" if (args.json or args.quiet) else "LogMonitor"

    try:
        with ObjectStore(args.cache) as object_store:
            if args.cache_max_size is not None:
                object_store.maximum_size = args.cache_max_size

            stage_timeout = args.stage_timeout
            debug_break = args.debug_break

            pipelines = manifest.depsolve(object_store, exports)
            total_steps = len(manifest.sources) + len(pipelines)
            monitor = osbuild.monitor.make(monitor_name, args.monitor_fd, total_steps)
            monitor.log(f"starting {args.manifest_path}", origin="osbuild.main_cli")

            r = manifest.build(
                object_store,
                pipelines,
                monitor,
                args.libdir,
                debug_break,
                in_vm=in_vm,
                stage_timeout=stage_timeout
            )
            if r.success:
                monitor.log(f"manifest {args.manifest_path} finished successfully\n", origin="osbuild.main_cli")
            else:
                # if we had monitor.error() we could use that here
                monitor.log(f"manifest {args.manifest_path} failed\n", origin="osbuild.main_cli")

            if r.success and exports:
                for pid in exports:
                    export(pid, output_directory, object_store, manifest)

            if args.json:
                json.dump(fmt.output(manifest, r, object_store), sys.stdout)
                sys.stdout.write("\n")
            elif not args.quiet:
                if r.success:
                    print("\nPipelines")
                    for name, pl in manifest.pipelines.items():
                        print(f"  {name + ':': <10}\t{pl.id}")

                    print("\nCheckpoints")
                    for m in marked:
                        if object_store.contains(m, only_cached=True):
                            print(f"  {m}: cached")
                        else:
                            print(f"  {m}: {vt.reset}{vt.bold}{vt.red}not cached{vt.reset}")
                else:
                    print(f"{vt.reset}{vt.bold}{vt.red}Failed{vt.reset}")

            return 0 if r.success else 1

    except KeyboardInterrupt:
        print()
        print(f"{vt.reset}{vt.bold}{vt.red}Aborted{vt.reset}")
        return 130



#!/usr/bin/python3
from unittest import mock

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.systemd-repart.create"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({"path": "image.raw"}, "'size' is a required property"),
    ({"size": "1G"}, "'path' is a required property"),
    ({"path": "image.raw", "size": "1G", "seed": 1}, "1 is not of type"),
    ({"path": 1, "size": "1G", "seed": "random"}, "1 is not of type"),
    ({"path": "image.raw", "size": 1, "seed": "random"}, "1 is not of type"),
    ({"path": "image.raw", "size": "1G", "sector-size": "512"}, "'512' is not of type"),
    ({"path": "image.raw", "size": "1G", "defer-partitions-factory-reset": "yes"}, "'yes' is not of type"),
    ({"path": "image.raw", "size": "1G", "split": "yes"}, "'yes' is not of type"),
    # good
    ({"path": "image.raw", "size": "1G"}, ""),
    ({"path": "image.raw", "size": "1G", "seed": "random"}, ""),
    ({"path": "image.raw", "size": "1G", "sector-size": 512}, ""),
    ({"path": "image.raw", "size": "1G", "sector-size": 4096}, ""),
    ({"path": "image.raw", "size": "1G", "defer-partitions-factory-reset": True}, ""),
    ({"path": "image.raw", "size": "1G", "defer-partitions-factory-reset": False}, ""),
    ({"path": "image.raw", "size": "1G", "split": True}, ""),
    ({"path": "image.raw", "size": "1G", "split": False}, ""),
])
def test_systemd_repart_create_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@mock.patch("subprocess.check_call")
@mock.patch("subprocess.run")
def test_systemd_repart_create_minimal(mock_run, mock_check_call, tmp_path, stage_module):
    fake_tree = tmp_path / "tree"
    fake_root = tmp_path / "root"

    options = {
        "path": "image.raw",
        "size": "1G",
    }

    stage_module.main(
        {
            "tree": fake_tree,
            "inputs": {"root-tree": {"path": str(fake_root)}},
            "options": options,
        },
        options,
    )

    mock_check_call.assert_any_call(["mount", "-t", "devtmpfs", "devtmpfs", "/dev/"])
    mock_run.assert_called_with([
        "systemd-repart",
        "--empty", "create",
        "--size", "1G",
        "--root", str(fake_root),
        "--seed", "random",
        str(fake_tree / "image.raw"),
    ], check=True)
    mock_check_call.assert_any_call(["umount", "/dev/"])


@mock.patch("subprocess.check_call")
@mock.patch("subprocess.run")
def test_systemd_repart_create_with_seed(mock_run, mock_check_call, tmp_path, stage_module):
    fake_tree = tmp_path / "tree"
    fake_root = tmp_path / "root"

    options = {
        "path": "disk.img",
        "size": "10G",
        "seed": "abc123",
    }

    stage_module.main(
        {
            "tree": fake_tree,
            "inputs": {"root-tree": {"path": str(fake_root)}},
            "options": options,
        },
        options,
    )

    mock_check_call.assert_any_call(["mount", "-t", "devtmpfs", "devtmpfs", "/dev/"])
    mock_run.assert_called_with([
        "systemd-repart",
        "--empty", "create",
        "--size", "10G",
        "--root", str(fake_root),
        "--seed", "abc123",
        str(fake_tree / "disk.img"),
    ], check=True)
    mock_check_call.assert_any_call(["umount", "/dev/"])


@mock.patch("subprocess.check_call")
@mock.patch("subprocess.run")
def test_systemd_repart_create_with_sector_size(mock_run, mock_check_call, tmp_path, stage_module):
    fake_tree = tmp_path / "tree"
    fake_root = tmp_path / "root"

    options = {
        "path": "image.raw",
        "size": "2G",
        "sector-size": 4096,
    }

    stage_module.main(
        {
            "tree": fake_tree,
            "inputs": {"root-tree": {"path": str(fake_root)}},
            "options": options,
        },
        options,
    )

    mock_check_call.assert_any_call(["mount", "-t", "devtmpfs", "devtmpfs", "/dev/"])
    mock_run.assert_called_with([
        "systemd-repart",
        "--empty", "create",
        "--size", "2G",
        "--root", str(fake_root),
        "--seed", "random",
        str(fake_tree / "image.raw"),
        "--sector-size", "4096",
    ], check=True)
    mock_check_call.assert_any_call(["umount", "/dev/"])


@mock.patch("subprocess.check_call")
@mock.patch("subprocess.run")
def test_systemd_repart_create_with_defer_partitions_factory_reset(mock_run, mock_check_call, tmp_path, stage_module):
    fake_tree = tmp_path / "tree"
    fake_root = tmp_path / "root"

    options = {
        "path": "image.raw",
        "size": "2G",
        "defer-partitions-factory-reset": True,
    }

    stage_module.main(
        {
            "tree": fake_tree,
            "inputs": {"root-tree": {"path": str(fake_root)}},
            "options": options,
        },
        options,
    )

    mock_check_call.assert_any_call(["mount", "-t", "devtmpfs", "devtmpfs", "/dev/"])
    mock_run.assert_called_with([
        "systemd-repart",
        "--empty", "create",
        "--size", "2G",
        "--root", str(fake_root),
        "--seed", "random",
        str(fake_tree / "image.raw"),
        "--defer-partitions-factory-reset", "yes",
    ], check=True)
    mock_check_call.assert_any_call(["umount", "/dev/"])


@mock.patch("subprocess.check_call")
@mock.patch("subprocess.run")
def test_systemd_repart_create_with_split(mock_run, mock_check_call, tmp_path, stage_module):
    fake_tree = tmp_path / "tree"
    fake_root = tmp_path / "root"

    options = {
        "path": "image.raw",
        "size": "2G",
        "split": True,
    }

    stage_module.main(
        {
            "tree": fake_tree,
            "inputs": {"root-tree": {"path": str(fake_root)}},
            "options": options,
        },
        options,
    )

    mock_check_call.assert_any_call(["mount", "-t", "devtmpfs", "devtmpfs", "/dev/"])
    mock_run.assert_called_with([
        "systemd-repart",
        "--empty", "create",
        "--size", "2G",
        "--root", str(fake_root),
        "--seed", "random",
        str(fake_tree / "image.raw"),
        "--split", "yes",
    ], check=True)
    mock_check_call.assert_any_call(["umount", "/dev/"])



#!/usr/bin/python3

import os.path
import subprocess
from unittest import mock

import pytest

from osbuild import testutil
from osbuild.testutil import has_executable, make_fake_input_tree

TEST_INPUT = [
    ({}, []),
    ({"compression": {"method": "lz4"}}, ["-comp", "lz4"]),
    ({"compression": {"method": "xz", "options": {"bcj": "x86"}}}, ["-comp", "xz", "-Xbcj", "x86"]),
    ({"exclude_paths": ["boot/.*", "root/.*"]}, ["-regex", "-e", "boot/.*", "root/.*"]),
    ({"source": "input://tree/"}, []),
    ({"source": "mount:///"}, []),
    ({"source": "mount://tree/"}, [])
]


STAGE_NAME = "org.osbuild.squashfs"


@pytest.mark.skipif(not has_executable("mksquashfs"), reason="no mksquashfs")
@pytest.mark.parametrize("test_options,expected", TEST_INPUT)
def test_squashfs_integration(tmp_path, stage_module, test_options, expected):  # pylint: disable=unused-argument
    fake_input_tree = make_fake_input_tree(tmp_path, {
        "/file-in-root.txt": "other content",
        "/subdir/file-in-subdir.txt": "subdir content",
    })
    inputs = {
        "tree": {
            "path": fake_input_tree,
        }
    }
    mounts = {
        "tree": {
            "path": fake_input_tree,
        }
    }
    paths = {
        "mounts": fake_input_tree
    }

    filename = "some-file.img"
    options = {
        "filename": filename,
    }
    options.update(test_options)

    stage_module.main({
        "inputs": inputs,
        "tree": tmp_path,
        "options": options,
        "mounts": mounts,
        "paths": paths
    })

    img_path = os.path.join(tmp_path, "some-file.img")
    assert os.path.exists(img_path)
    # validate the content
    output = subprocess.check_output([
        "unsquashfs", "-ls", img_path], encoding="utf-8")
    assert "subdir/file-in-subdir.txt\n" in output
    assert "file-in-root.txt\n" in output


@mock.patch("subprocess.run")
@pytest.mark.parametrize("test_options,expected", TEST_INPUT)
def test_squashfs(mock_run, tmp_path, stage_module, test_options, expected):
    fake_input_tree = make_fake_input_tree(tmp_path, {
        "/some-dir/some-file.txt": "content",
    })
    inputs = {
        "tree": {
            "path": fake_input_tree,
        }
    }
    mounts = {
        "tree": {
            "path": fake_input_tree,
        }
    }
    paths = {
        "mounts": fake_input_tree
    }

    filename = "some-file.img"
    options = {
        "filename": filename,
    }
    options.update(test_options)

    stage_module.main({
        "inputs": inputs,
        "tree": tmp_path,
        "options": options,
        "mounts": mounts,
        "paths": paths
    })

    expected = [
        "mksquashfs",
        f"{fake_input_tree}",
        f"{os.path.join(tmp_path, filename)}",
    ] + expected
    mock_run.assert_called_with(expected, check=True)


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({"extra": "option"}, "'extra' was unexpected"),
    ({"compression": {"method": "invalid"}}, "'invalid' is not one of ["),
    ({"compression": {}}, "'method' is a required property"),
    ({"compression": {"method": "xz", "options": {"bcj": "invalid"}}}, "'invalid' is not one of ["),
    ({"compression": {"method": "xz", "options": {"level": 9}}}, "Additional properties are not allowed"),
    # good
    ({"compression": {"method": "lz4"}}, ""),
    ({"exclude_paths": ["boot/", "root/"]}, "")
])
def test_schema_validation_squashfs(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
            "filename": "some-filename.img",
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)



#!/usr/bin/python3

import pathlib
from unittest import mock

STAGE_NAME = "org.osbuild.systemd-sysusers"


@mock.patch("subprocess.run")
def test_systemd_sysusers(mock_run, tmp_path, stage_module):
    stage_module.main(tmp_path)

    expected = [
        "systemd-sysusers",
        "--root",
        pathlib.Path(tmp_path),
    ]
    mock_run.assert_called_with(expected, check=True)
